package com.humuson.amc.client;

import com.humuson.amc.client.constant.GrantType;
import com.humuson.amc.client.constant.SessionStrategyType;
import com.humuson.amc.client.context.ClientContext;
import com.humuson.amc.client.context.Session;
import com.humuson.amc.client.context.SessionStore;
import com.humuson.amc.client.controller.AppController;
import com.humuson.amc.client.controller.AudienceController;
import com.humuson.amc.client.controller.ContactController;
import com.humuson.amc.client.controller.ContentController;
import com.humuson.amc.client.controller.ServiceController;
import com.humuson.amc.client.controller.SessionController;
import com.humuson.amc.client.controller.SharingController;
import com.humuson.amc.client.controller.SiteController;
import com.humuson.amc.client.controller.SubjectInfoController;
import com.humuson.amc.client.controller.TargetController;
import com.humuson.amc.client.controller.TrackingLogController;
import com.humuson.amc.client.controller.UserController;
import com.humuson.amc.client.exception.InvalidSessionException;
import com.humuson.amc.client.http.HttpComunicator;
import com.humuson.amc.client.http.NonSecurityHttpComunicator;
import com.humuson.amc.client.http.SecurityAwareHttpComunicator;
import com.humuson.amc.client.info.Optional;
import com.humuson.amc.client.info.RequiredOptional;
import com.humuson.amc.client.info.UriInfo;
import com.humuson.amc.client.model.Authentication;
import com.humuson.amc.client.model.Response;

public class AmcClient {
		
	private ClientContext context;
	
	public ContentController content;
	public AudienceController audience;
	public SiteController site;
	public AppController app;
	public ContactController contact;
	public UserController user;
	public SessionController session;
	public SharingController sharing;
	public TargetController target;
	public SubjectInfoController subject;
	public TrackingLogController trk;
	public ServiceController service;
	
	private HttpComunicator defaultDefaultHttpComunicator = null;
	private SessionStore sessionStore;
	
	private AmcClient(Builder builder) {
		
		if ( !builder.useSecurity ) {
			defaultDefaultHttpComunicator = new NonSecurityHttpComunicator(this);
		} else {
			defaultDefaultHttpComunicator = new SecurityAwareHttpComunicator(this);
		}
		
		content		= new ContentController(this);
		audience	= new AudienceController(this);
		site		= new SiteController(this);
		app			= new AppController(this);
		contact		= new ContactController(this);
		user		= new UserController(this);
		session		= new SessionController(this);
		sharing		= new SharingController(this);
		target		= new TargetController(this);
		subject		= new SubjectInfoController(this);
		trk			= new TrackingLogController(this);
		service		= new ServiceController(this);
		
		Authentication auth	= builder.getAuthentication();
				
		context = new ClientContext(builder.amcUriInfo);
		context.setDefaultAuthentication(auth);
		
		if ( builder.sessionStrategyType != null ) {
			sessionStore = new SessionStore(this, builder.sessionStrategyType);
		} else {
			sessionStore = new SessionStore(this);
		}
		
		sessionStore.clearSession();
		sessionStore.getSession().setAuthentication(auth).setSessionId(builder.sessionId);
	}
	
	public Authentication authorize() {
		return defaultDefaultHttpComunicator.authorize();
	}
	
	public Authentication authorize(Integer timeout) {
		return defaultDefaultHttpComunicator.authorize(timeout);
	}
	
	public void makeSession() {
		
		Response<String> result = session.getSessionId();
		if ( !result.isSucccess() ) {
			throw new InvalidSessionException("fail make session. result:"+result.toString());
		}
		
		String sessionId = result.get();
		sessionStore.getSession().setSessionId(sessionId);
	}
	
	public void makeSession(Integer timeout) {
		
		Response<String> result = session.getSessionId(timeout);
		if ( !result.isSucccess() ) {
			throw new InvalidSessionException("fail make session. result:"+result.toString());
		}
		
		String sessionId = result.get();
		sessionStore.getSession().setSessionId(sessionId);
	}
	
	public Authentication authorizeAndMakeSession() {
		Authentication auth = authorize();
		makeSession();
		return auth;
	}
	
	public Authentication authorizeAndMakeSession(Integer timeout) {
		Authentication auth = authorize(timeout);
		makeSession(timeout);
		return auth;
	}
	
	public HttpComunicator getDefaultHttpComunicator() {
		return defaultDefaultHttpComunicator;
	}
	
	public String getTokenUri() {
		return context.getUriInfo().getTokenUri();
	}
	
	public String makeFullPath(String path) {
		return context.getUriInfo().getAmcUri()+path;
	}
	
	public static Builder newBuilder() {
		return new Builder();
	}
	
	public ClientContext getClientContext() {
		return context;
	}
	
	public Session getSession() {
		return sessionStore.getSession();
	}
	
	public Authentication getAuthentication() {
		return getSession().getAuthentication();
	}
	
	public void setToken(String token) {
		getSession().getAuthentication().setToken(token);
	}
	
	public void setRefreshToken(String refreshToken) {
		getSession().getAuthentication().setRefreshToken(refreshToken);
	}
	
	public void setSessionId(String sessionId) {
		getSession().setSessionId(sessionId);
	}
		
	public void setUserIdAndPassword(String id, String password) {
		getSession().getAuthentication().setUserId(id).setUserPassword(password);
	}
	
	public void clearUser() {
		getSession().getAuthentication().setUserId(null).setUserPassword(null);
	}
	
	
	public void setUserId(String id) {
		getSession().getAuthentication().setUserId(id);
	}
	
	public void setUserPassword(String password) {
		getSession().getAuthentication().setUserPassword(password);
	}
	
	public static class Builder {
		
		@RequiredOptional
		private String userId;
		
		@RequiredOptional
		private String userPassword;
		
		@RequiredOptional
		private String serviceId;
		
		@RequiredOptional
		private String servicePassword;
		
		@RequiredOptional
		private String token;
		
		@RequiredOptional
		private String refreshToken;
		
		@RequiredOptional
		private UriInfo amcUriInfo;
		
		@Optional
		private boolean useSecurity = true;
		
		@Optional
		private GrantType tokenGrantType;
		
		@Optional
		private SessionStrategyType sessionStrategyType;
		
		@Optional
		private String sessionId;
		
		public Builder setUserId(String id) {
			this.userId 			= id;
			return this;
		}
		
		public Builder setUserPassword(String password) {
			this.userPassword		= password;
			return this;
		}
		
		public Builder setUserIdAndPassword(String id, String password) {
			this.userId				= id;
			this.userPassword		= password;
			return this;
		}

		public Builder setServiceId(String id) {
			this.serviceId 			= id;
			return this;
		}
		
		public Builder setServicePassword(String password) {
			this.servicePassword 	= password;
			return this;
		}
		
		public Builder setServiceIdAndPassword(String id, String password) {
			this.serviceId 			= id;
			this.servicePassword	= password;
			return this;
		}

		public Builder setToken(String token) {
			this.token = token;
			return this;
		}

		public Builder setRefreshToken(String refreshToken) {
			this.refreshToken = refreshToken;
			return this;
		}
		
		public Builder setAmcUriInfo(UriInfo amcUriInfo) {
			
			if ( amcUriInfo != null ) {
				this.amcUriInfo = amcUriInfo;
			}
			
			return this;
		}

		public Builder setBaseUri(String baseUri) {
			
			if ( amcUriInfo == null ) {
				amcUriInfo = new UriInfo(); 
			}
			
			amcUriInfo.setAmcUri(baseUri);
			return this;
		}
		
		public Builder setTokenUri(String tokenUri) {
			
			if ( amcUriInfo == null ) {
				amcUriInfo = new UriInfo(); 
			}
			
			amcUriInfo.setTokenUri(tokenUri);
			return this;
		}
		
		
		public Builder setSessionStrategy(SessionStrategyType sessionStrategyType) {
			this.sessionStrategyType = sessionStrategyType;
			return this;
		}
		
		public Builder setTokenGrantType(GrantType grantType) {
			this.tokenGrantType = grantType;
			return this;
		}
		
		public Builder setUseSecurity(boolean useSecurity) {
			this.useSecurity = useSecurity;
			return this;
		}
		
		public Builder setSessionId(String sessionId) {
			this.sessionId = sessionId;
			return this;
		}
		
		public Authentication getAuthentication() {
			
			Authentication authentication = new Authentication();
			
			authentication.setUserId(userId);
			authentication.setUserPassword(userPassword);
			authentication.setServiceId(serviceId);
			authentication.setServicePassword(servicePassword);
			authentication.setToken(token);
			authentication.setRefreshToken(refreshToken);
			authentication.setGrantType(tokenGrantType);
			
			return authentication;
			
		}

		public AmcClient build() {
			return new AmcClient(this);
		}
	}
}
