package com.humuson.amc.client.http;

import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.Map;

import com.humuson.amc.client.AmcClient;
import com.humuson.amc.client.constant.GrantType;
import com.humuson.amc.client.context.Session;
import com.humuson.amc.client.exception.AuthenticationException;
import com.humuson.amc.client.model.Authentication;
import com.humuson.amc.client.model.Request;
import com.humuson.amc.client.model.Response;
import com.humuson.amc.client.util.HttpUtil;
import com.humuson.amc.client.util.ReflectionUtil;
import com.humuson.amc.client.util.StringUtils;
import com.humuson.amc.client.util.TypeUtil;

public class SecurityAwareHttpComunicator implements HttpComunicator {
	
	//private static final Logger log = LoggerFactory.getLogger(SecurityAwareHttpComunicator.class);
	
	private static final String KEY_PREV_AUTH = "prev_auth";
		
	AmcClient client;
	
	public SecurityAwareHttpComunicator(AmcClient client) {
		this.client = client;
	}
	
	private boolean preAuthorize(Session session) {

		Authentication currAuth = session.getAuthentication();
		
		Object o = session.getInfos().get(KEY_PREV_AUTH);
		session.getInfos().put(KEY_PREV_AUTH, currAuth.copy());
		
		//이전 authenticaiton 정보와 비교하여 달라졌다면 fals반환
		if ( o != null && !currAuth.isSameWithoutToken( (Authentication)o ) ) {
			session.clear();
			return false;
		}
		
		if ( session.getSessionId() != null ) {
			return true;
		}
		
		if ( currAuth.isAuthenticated() ) {
			return true;
		} else {
			return false;
		}
	}
	
	private Map<String, String> makeAuthorizeHeader( Map<String, String> header ) {
		
		Map<String,String> newHeader = new HashMap<String,String>();
		
		Map<String, String> authorizeHeader = client.getSession().getAuthorizeHeader();
		
		if ( header != null ) {
			newHeader.putAll(header);	
		}
		
		if ( authorizeHeader != null ) {
			newHeader.putAll(authorizeHeader);
		}
		
		return newHeader;
	}
	
	@Override
	public Authentication authorize() {
		return authorize(null);
	}
	
	private void processAuthenticationException(String msg, Authentication authentication, Map<String, String> clientHeader, Map<String, Object> params, Exception e) {
		
		String errorMsg = new StringBuilder()
			.append("msg:").append(msg)
			.append(", request info : [url:").append(client.getTokenUri())
			.append(", header:").append(ReflectionUtil.toJson(clientHeader))
			.append(", params:").append(ReflectionUtil.toJson(params))
			.append(", authentication:").append(authentication.toString())
			.append("]").toString();
		
		authentication.setToken(null);
		authentication.setRefreshToken(null);
		
		throw new AuthenticationException(errorMsg,e);
	}
	
	@Override
	public Authentication authorize(Integer timeout) {
				
		Session session = client.getSession();
		
		session.clear();
		
		Authentication authentication = session.getAuthentication();
		
		Map<String, String> clientHeader = authentication.getClientAuthorizationHeader();
		
		Map<String, Object> params = new HashMap<String, Object>();
		
		GrantType grantType = authentication.findRequestTokenGrantType();
		
		params.put("grant_type", grantType.getCode());
		
		if ( grantType == GrantType.Password ) {
			
			params.put("username", authentication.getUserId());
			params.put("password", authentication.getUserPassword());
			
		} else if ( grantType == GrantType.ClientUserCredentials ) {
			
			params.put("user_name", authentication.getUserId());
			
		} else if ( grantType == GrantType.RefreshToken ) {
			
			params.put("refresh_token", authentication.getRefreshToken());
			
		}
		Response<Map<String, String>> result = null;
		
		try {
			
			result = HttpUtil.executePost(client.getTokenUri(), clientHeader, params, timeout, TypeUtil.getStringMapType(), 0);
			
			
		} catch (Exception e) {
			
			processAuthenticationException("Failed to get token",authentication,clientHeader, params,e);
		}
		
		if ( !result.isSucccess() ) {
			processAuthenticationException(result.toString(),authentication,clientHeader, params, null);
		}
		
		String accessToken = result.get().get("access_token");
		String refreshToken = result.get().get("refresh_token");
		
		if ( StringUtils.isEmpty(accessToken) ) {
			processAuthenticationException("access token not found",authentication,clientHeader, params, null);
		}
		
		authentication.setToken(accessToken);
		authentication.setRefreshToken(refreshToken);
		
		return authentication;
	}
	
	@Override
	public <T> Response<T> executePost(String path, Request request, Type type) {
		
		Map<String,Object> params = null;
		Integer timeout = null;
		Integer retry = null;
		
		if ( request != null ) {
			params	= request.getParams();
			timeout	= request.getTimeout();
			retry	= request.getRetryCount();
		}
		
		return executePost(path, request.getHttpHeaders(), params, timeout, type, retry);
	}
	
	@Override
	public <T> Response<T> executePost(
			String path, Map<String,String> headers,
			Map<String,Object> params, Integer timeout,
			Type type, Integer retry) {
						
		return execute(path, headers, timeout, new Action(path, params, timeout, type, retry) {
			
			@SuppressWarnings("hiding")
			@Override
			public <T> Response<T> execute(Map<String,String> headers) {
				String uri = client.makeFullPath(path);
				return HttpUtil.executePost(uri, headers, params, timeout, type, retry);
			}
			
		});
	}

	@Override
	public <T> Response<T> executeGet(String path, Request request, Type type) {
		
		Map<String,Object> params = null;
		Integer timeout = null;
		Integer retry	= null;
		
		if ( request != null ) {
			params	= request.getParams();
			timeout	= request.getTimeout();
			retry	= request.getRetryCount();
		}
		
		return executeGet(path, request.getHttpHeaders(), params, timeout, type, retry);
	}

	@Override
	public <T> Response<T> executeGet(
			String path, Map<String, String> headers,
			Map<String, Object> params, Integer timeout,
			Type type, Integer retry) {
		
		return execute(path, headers, timeout, new Action(path, params, timeout, type, retry) {
			
			@SuppressWarnings("hiding")
			@Override
			public <T> Response<T> execute(Map<String,String> headers) {
				String uri = client.makeFullPath(path);
				return HttpUtil.executeGet(uri, headers, params, timeout, type, retry);
			}
			
		});
	}

	@Override
	public <T> Response<T> executePut(String path, Request request, Type type) {
		
		Map<String,Object> params = null;
		Integer timeout = null;
		Integer retry	= null;
		
		if ( request != null ) {
			params	= request.getParams();
			timeout	= request.getTimeout();
			retry	= request.getRetryCount();
		}
		
		return executePut(path, request.getHttpHeaders(), params, timeout, type, retry);
	}

	@Override
	public <T> Response<T> executePut(
			String path, Map<String, String> headers,
			Map<String, Object> params, Integer timeout,
			Type type, Integer retry) {

		return execute(path, headers, timeout, new Action(path, params, timeout, type, retry) {
			
			@SuppressWarnings("hiding")
			@Override
			public <T> Response<T> execute(Map<String,String> headers) {
				String uri = client.makeFullPath(path);
				return HttpUtil.executePut(uri, headers, params, timeout, type, retry);
			}
			
		});
	}

	@Override
	public <T> Response<T> executeDelete(String path, Request request, Type type) {
		
		Map<String,Object> params = null;
		Integer timeout = null;
		Integer retry	= null;
		
		if ( request != null ) {
			params	= request.getParams();
			timeout	= request.getTimeout();
			retry	= request.getRetryCount();
		}
		
		return executeDelete(path, request.getHttpHeaders(), params, timeout, type, retry);
	}

	@Override
	public <T> Response<T> executeDelete(
			String path, Map<String, String> headers,
			Map<String, Object> params, Integer timeout,
			Type type, Integer retry) {

		return execute(path, headers, timeout, new Action(path, params, timeout, type, retry) {
			
			@SuppressWarnings("hiding")
			@Override
			public <T> Response<T> execute(Map<String,String> headers) {
				String uri = client.makeFullPath(path);
				return HttpUtil.executeDelete(uri, headers, params, timeout, type, retry);
			}
			
		});
	}
	
	
	public <T> Response<T> execute(String path, Map<String,String> headers, Integer timeout, Action action) {
				
		boolean checkCallAuthorize = false;
		
		boolean isAlreadyValidAuth = false;
		
		Map<String,String> reqHeader = null;
		
		Session session = client.getSession();
		
		try {
			isAlreadyValidAuth = preAuthorize(session);	
		} catch ( Exception e ) {
			//not throw
			//log.warn("invalid token",e);
			e.printStackTrace();
		}
		
		if ( !isAlreadyValidAuth ) {
			
			authorize(timeout);
			
			checkCallAuthorize = true;
		}
		
		reqHeader = makeAuthorizeHeader(headers);
		
		Response<T> response =  action.execute(reqHeader);
		
		/**
		 * token 만료, 인증 실패 등의 이유.
		 * 아래 로직은 토큰 만료의 경우를 대비하여 추가된 로직.
		 * 2017.09.25 login redirect를 고려한 로직 추가 (status 302)
		 */
		if ( !checkCallAuthorize && ( response.getHttpStatusCode() == 401 || response.getHttpStatusCode() == 302 ) ) {
			
			authorize(timeout);
			
			reqHeader = makeAuthorizeHeader(headers);
			
			response = action.execute(reqHeader);
			
		}
		
		session.setSessionId(response.getHeaders());
						
		return response;
	}
	
	private static abstract class Action {
		
		String path;
		Map<String, Object> params;
		Integer timeout;
		Type type;
		Integer retry;
		
		public Action(String path,Map<String, Object> params,
				Integer timeout, Type type, Integer retry) {
			this.path		= path;
			this.params		= params;
			this.timeout	= timeout;
			this.type		= type;
			this.retry		= retry;
		}
		public abstract <T> Response<T> execute(Map<String,String> headers);
	}
}