package com.humuson.amc.client.util;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Type;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.NameValuePair;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.utils.URLEncodedUtils;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.AbstractHttpMessage;
import org.apache.http.message.BasicNameValuePair;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.humuson.amc.client.http.RetryStrategyHandler;
import com.humuson.amc.client.model.Response;
import com.humuson.amc.client.model.response.ApiErrorForm;

public class HttpUtil {
	
	private static Gson gson;
	
	public static final int TIMEOUT = 30000;
	
	public static final RequestConfig DEFAULT_CONFIG = RequestConfig.custom()
															.setSocketTimeout(TIMEOUT)
												            .setConnectTimeout(TIMEOUT)
												            .setConnectionRequestTimeout(TIMEOUT)
												            .build();
	
	public static RequestConfig makeRequestConfig(int timeout) {
		return RequestConfig.custom()
					.setSocketTimeout(timeout)
			        .setConnectTimeout(timeout)
			        .setConnectionRequestTimeout(timeout)
			        .build();
	}
	
	static {
		gson = new GsonBuilder()
				.disableHtmlEscaping()
				.registerTypeAdapter(Date.class, new JsonDateAdapter()).create();
	}
	
	@SuppressWarnings("unchecked")
	public static <T> Response<T> execute(HttpRequestBase request, Map<String,String> headers, Type type, Integer timeout, Integer retry) {
		
		Response<T> result = new Response<T>();
		
		try {
			
			HttpClientBuilder httpBuilder = HttpClientBuilder.create();
			
			if ( retry != null && retry > 0 ) {
				httpBuilder.setServiceUnavailableRetryStrategy(new RetryStrategyHandler(retry));
			}
							
			RequestConfig config = (timeout != null && timeout != 0)?makeRequestConfig(timeout):DEFAULT_CONFIG;
			
			CloseableHttpClient client = httpBuilder.setDefaultRequestConfig(config).build();
			
			addHeader(request, headers);
			
			addAcceptHeader(request);
			
			HttpResponse response = client.execute(request);
			BufferedReader rd = new BufferedReader(
	                new InputStreamReader(response.getEntity().getContent()));

			StringBuilder resultBuilder = new StringBuilder();
			String line = "";
			
			while ((line = rd.readLine()) != null) {
				resultBuilder.append(line);
			}
			
			int statusCode = response.getStatusLine().getStatusCode();
			
			result.setHttpStatusCode(statusCode);
			
			result.setHeaders(toMap(response.getAllHeaders()));
			
			if ( statusCode != HttpStatus.SC_OK ) {
				
				ApiErrorForm error = gson.fromJson(resultBuilder.toString(), ApiErrorForm.class);
				
				if ( error != null && error.getResult() != null ) {
					
					result.setErrorCode(error.getResult().getCode());
					
					result.setMsg(error.getResult().getMsg());
					
				} else {
					
					result.setMsg(resultBuilder.toString());
				}
								
			} else {
				
				if ( type == String.class ) {
					result.set((T)resultBuilder.toString());
				} else {
					result.set( (T) gson.fromJson(resultBuilder.toString(), type));
				}
			}
			
		} catch (Exception e) {
			
			result.setHttpStatusCode(Response.ERROR);
			
			if ( e instanceof SocketTimeoutException ) {
				result.setMsg("Read timed out");
			} else {
				result.setMsg(e.getMessage());
			}
			
		} finally {
			request.releaseConnection();
		}
		
		return result;
		
	}
	
	public static <T> Response<T> executeGet(String url, Map<String,String> headers, Map<String,Object> params, Integer timeout, Type type, Integer retry) {
		
		HttpGet request = new HttpGet(makeParamUrl(url, params));
				
		Response<T> result = execute(request, headers, type, timeout, retry);
		
		return result;
		
	}
		
	public static <T> Response<T> executePost(String url, Map<String,String> headers, Map<String,Object> params, Integer timeout, Type type, Integer retry) {
		
		HttpPost request = new HttpPost(url);
		
		request.setEntity(toFormEntity(headers, params));
						
		Response<T> result = execute(request, headers, type, timeout, retry);
		
		return result;
		
	}
	
	public static <T> Response<T> executePut(String url, Map<String,String> headers, Map<String,Object> params, Integer timeout, Type type, Integer retry) {
		
		HttpPut request = new HttpPut(url);
		
		request.setEntity(toFormEntity(headers, params));
						
		Response<T> result = execute(request, headers, type, timeout, retry);
		
		return result;
		
	}
	
	public static <T> Response<T> executeDelete(String url, Map<String,String> headers, Map<String,Object> params, Integer timeout, Type type, Integer retry) {
		
		HttpDelete request = new HttpDelete(makeParamUrl(url, params));
						
		Response<T> result = execute(request, headers, type, timeout, retry);
		
		return result;
		
	}
	
	private static String makeParamUrl(String url, Map<String,Object> params ) {
		
		if ( params == null ) {
			
			return url;
			
		} else {
			
			List<NameValuePair> paramList = toParams(params);
			
			return url+"?"+URLEncodedUtils.format(paramList, "UTF-8");
		}
	}
	
	public static String toListParam(Collection<?> listParam) {
		StringBuilder builder = new StringBuilder();
		for ( Object param : listParam ) {
			builder.append(param).append(",");
		}
		
		if ( builder.length() == 0 ) {
			return null;
		}
		
		return builder.substring(0, builder.length()-1);
	}
	
	
	private static void addHeader(AbstractHttpMessage httpMessage, Map<String, String> headers) {
		
		if ( httpMessage == null || headers == null || headers.size() == 0 ) {
			return;
		}
		
		for ( String header : headers.keySet() ) {
			httpMessage.addHeader(header, headers.get(header));
		}
	}
	
	private static void addAcceptHeader(AbstractHttpMessage httpMessage) {
		
		httpMessage.addHeader("Accept", "application/json");
	}
	
	
	private static Map<String, String> toMap(Header[] headers) {
		
		if ( headers == null ) {
			return null;
		}
		
		Map<String, String> headerMap = new HashMap<String, String>();
		
		for ( Header header : headers ) {
			headerMap.put(header.getName(), header.getValue());
		}
		
		return headerMap;
	}
	
	@SuppressWarnings("rawtypes")
	private static List<NameValuePair> toParams(Map<String,Object> params) {
		if ( params == null ) {
			return null;
		}
		
		List<NameValuePair> urlParameters = new ArrayList<NameValuePair>();
		
		for ( String name : params.keySet() ) {
			
			Object value = params.get(name);
			
			if ( value == null ) {
				continue;
			}
			
			if ( value instanceof Collection ) {
				urlParameters.add(new BasicNameValuePair(name, toListParam( (Collection)value )));
			} else {
				urlParameters.add(new BasicNameValuePair(name, value.toString()));
			}
		}
		
		return urlParameters;
		
	}
	
	private static HttpEntity toFormEntity(Map<String,String> headers, Map<String,Object> params) {
		
		if ( params == null ) {
			return null;
		}
		
		try {
			
			if ( headers != null && "application/json".equals(headers.get("Content-Type")) ) {
				
				return new StringEntity(gson.toJson(params), ContentType.APPLICATION_JSON);
				
			} else {
				
				List<NameValuePair> urlParameters = toParams(params);
				
				return new UrlEncodedFormEntity(urlParameters, "UTF-8");
			}
			
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}
		
	}
}
